/*
		This shader will perform the first subdivision step
		on the GPU using a weighting texture.
		This shader can be used with for normals and vertices by attaching two
		color-textures to the framebuffer.
		The weighting will be the same for normals and vertices based on the paper
		"Subdivision Shading" by Tamy Boubekeur and Marc Alexa
		
		Matthias Hollnder 2008/2009
*/

// --------------------------------------------------
// shader definition
// --------------------------------------------------
#version 120
#extension GL_EXT_gpu_shader4 : enable

// --------------------------------------------------
// uniform variables:
// --------------------------------------------------
uniform int				CalculationSamplerSize;
uniform sampler2D		CalculationSampler;					// this is the texture used for performing the 1st Subdivision on the GPU
uniform int				iVertexStride;						// how many vec4 are for each vertex

uniform int				iInputSamplerSize;					// The size of the InputSamplers
uniform sampler2D		InputSamplerVertices;				// vertices
uniform sampler2D		InputSamplerNormals;				// normals

uniform int				iOutputSamplerSize;

// --------------------------------------------------
// varying variables
// --------------------------------------------------
varying vec3			vVarColorVertex;
varying vec3			vVarColorNormal;

// --------------------------------------------------
// function to convert 1D-index to 2D-index
// --------------------------------------------------
vec2 Convert1DIndexTo2DIndex(in unsigned int uiVertexIndex, in int iWrapSize)
{
		float fWrapRec = 1.0 / float(iWrapSize);
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiVertexIndex / uiWrap;
		unsigned int uiX = uiVertexIndex - (uiY* uiWrap);
		
		float fX = float(uiX) * fWrapRec;
		float fY = float(uiY) * fWrapRec;
		
		return vec2(fX, fY);
}

// --------------------------------------------------
// function to convert output position
// --------------------------------------------------
vec2 ConvertOutputPosition(in unsigned int uiIndex, in int iWrapSize)
{
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiIndex / uiWrap;
		
		float fX = float(uiIndex - (uiY*uiWrap));
		float fY = float(uiY);
		
		return vec2(fX, fY);
}

// --------------------------------------------------
// Apply 1st Subdivision
// --------------------------------------------------
void Apply1stSubdivision(in unsigned int uiVertexID, out vec3 vVertex, out vec3 vNormal)
{
		vec3 vReturnVertex = vec3( 0.0, 0.0, 0.0 );
		vec3 vReturnNormal = vec3( 0.0, 0.0, 0.0 );
		vec2 vTexCoord = vec2( 0.0, 0.0);
		vec4 vWeightAndIndex = vec4( 0.0, 0.0, 0.0, 0.0 );
		unsigned int  uiIndex = 0u;
		float fWeight = 0.0;
		
		unsigned int uiVertexIndex = unsigned int( uiVertexID ) * unsigned int( iVertexStride );		// Position of Weight/Index-struct
		
		// get the vertices and weights:
		for ( unsigned int i=0u; i < unsigned int( iVertexStride ); ++i ) {							// Process One-Ring-Neighbourhood (and self)
			vTexCoord = Convert1DIndexTo2DIndex( uiVertexIndex+i, CalculationSamplerSize );		// Get TexCoord of struct
			vWeightAndIndex = texture2D( CalculationSampler, vTexCoord );							// Get struct
			
			// first vertex and normal:
			fWeight = vWeightAndIndex.x;
			if ( fWeight < 0.0 ) { break; }
			uiIndex = unsigned int(vWeightAndIndex.y);
			vTexCoord = Convert1DIndexTo2DIndex( uiIndex, iInputSamplerSize );
			vReturnVertex += texture2D( InputSamplerVertices, vTexCoord ).xyz * fWeight;
			vReturnNormal += texture2D( InputSamplerNormals, vTexCoord ).xyz * fWeight;
			
			// second vertex and normal:
			fWeight = vWeightAndIndex.z;
			if ( fWeight < 0.0 ) { break; }
			uiIndex = unsigned int(vWeightAndIndex.w);
			vTexCoord = Convert1DIndexTo2DIndex(uiIndex, iInputSamplerSize);
			vReturnVertex += texture2D( InputSamplerVertices, vTexCoord ).xyz * fWeight;
			vReturnNormal += texture2D( InputSamplerNormals, vTexCoord ).xyz * fWeight;
		}
		
		vVertex = vReturnVertex;
		vNormal = vReturnNormal;
}

// --------------------------------------------------
// Vertex-Shader
// --------------------------------------------------
void main() {
		// the point we instance is 0,0,0 so the + does not actually add something
		// but this way it works without warnings
		unsigned int uiVertexID = unsigned int ( gl_InstanceID ) + unsigned int( gl_Vertex.x );
		vec3 vNormal = vec3( 0.0, 0.0, 0.0 );
		vec3 vVertex = vec3( 0.0, 0.0, 0.0 );
		Apply1stSubdivision( uiVertexID, vVertex, vNormal );
		
		vVarColorVertex = vVertex;												// color1 is the subdivision vertex
		vVarColorNormal = vNormal;												// color2 is the subdivision normal

		vec2 vPos2D = ConvertOutputPosition( uiVertexID, iOutputSamplerSize );	// The position to write to
		vec4 vPos = vec4( vPos2D.xy, 0.0, 1.0 );								// z=0.0, w=1.0
		
		gl_Position = ( gl_ModelViewProjectionMatrix * vPos );
}
